"""Configuration.from_env() tests."""

from pytest import mark, raises


def test(config):
    config.from_env("CONFIG_TEST_ENV")
    assert config() == "test-value"


def test_with_children(config):
    config.section1.value1.from_env("CONFIG_TEST_ENV")

    assert config() == {"section1": {"value1": "test-value"}}
    assert config.section1() == {"value1": "test-value"}
    assert config.section1.value1() == "test-value"


def test_default(config):
    config.from_env("UNDEFINED_ENV", "default-value")
    assert config() == "default-value"


def test_default_none(config):
    config.from_env("UNDEFINED_ENV")
    assert config() is None


def test_option_default_none(config):
    config.option.from_env("UNDEFINED_ENV")
    assert config.option() is None


def test_as_(config):
    config.from_env("CONFIG_INT", as_=int)
    assert config() == 42
    assert isinstance(config(), int)


def test_as__default(config):
    config.from_env("UNDEFINED", as_=int, default="33")
    assert config() == 33
    assert isinstance(config(), int)


def test_as__undefined_required(config):
    with raises(ValueError):
        config.from_env("UNDEFINED", as_=int, required=True)
    assert config() == {}


def test_as__defined_empty(config):
    with raises(ValueError):
        config.from_env("EMPTY", as_=int)
    assert config() == {}


def test_option_as_(config):
    config.option.from_env("CONFIG_INT", as_=int)
    assert config.option() == 42
    assert isinstance(config.option(), int)


def test_option_as__default(config):
    config.option.from_env("UNDEFINED", as_=int, default="33")
    assert config.option() == 33
    assert isinstance(config.option(), int)


def test_option_as__undefined_required(config):
    with raises(ValueError):
        config.option.from_env("UNDEFINED", as_=int, required=True)
    assert config.option() is None


def test_option_as__defined_empty(config):
    with raises(ValueError):
        config.option.from_env("EMPTY", as_=int)
    assert config.option() is None


@mark.parametrize("config_type", ["strict"])
def test_undefined_in_strict_mode(config):
    with raises(ValueError):
        config.from_env("UNDEFINED_ENV")


@mark.parametrize("config_type", ["strict"])
def test_option_undefined_in_strict_mode(config):
    with raises(ValueError):
        config.option.from_env("UNDEFINED_ENV")


def test_undefined_in_strict_mode_with_default(config):
    config.from_env("UNDEFINED_ENV", "default-value")
    assert config() == "default-value"


@mark.parametrize("config_type", ["strict"])
def test_option_undefined_in_strict_mode_with_default(config):
    config.option.from_env("UNDEFINED_ENV", "default-value")
    assert config.option() == "default-value"


def test_required_undefined(config):
    with raises(ValueError):
        config.from_env("UNDEFINED_ENV", required=True)


def test_required_undefined_with_default(config):
    config.from_env("UNDEFINED_ENV", default="default-value", required=True)
    assert config() == "default-value"


def test_option_required_undefined(config):
    with raises(ValueError):
        config.option.from_env("UNDEFINED_ENV", required=True)


def test_option_required_undefined_with_default(config):
    config.option.from_env("UNDEFINED_ENV", default="default-value", required=True)
    assert config.option() == "default-value"


@mark.parametrize("config_type", ["strict"])
def test_not_required_undefined_in_strict_mode(config):
    config.from_env("UNDEFINED_ENV", required=False)
    assert config() is None


@mark.parametrize("config_type", ["strict"])
def test_option_not_required_undefined_in_strict_mode(config):
    config.option.from_env("UNDEFINED_ENV", required=False)
    assert config.option() is None


@mark.parametrize("config_type", ["strict"])
def test_not_required_undefined_with_default_in_strict_mode(config):
    config.from_env("UNDEFINED_ENV", default="default-value", required=False)
    assert config() == "default-value"


@mark.parametrize("config_type", ["strict"])
def test_option_not_required_undefined_with_default_in_strict_mode(config):
    config.option.from_env("UNDEFINED_ENV", default="default-value", required=False)
    assert config.option() == "default-value"



"""Configuration provider tests."""

from dependency_injector import providers


def test_traverse():
    config = providers.Configuration(default={"option1": {"option2": "option2"}})
    option1 = config.option1
    option2 = config.option1.option2
    option3 = config.option1[config.option1.option2]

    all_providers = list(config.traverse())

    assert len(all_providers) == 3
    assert option1 in all_providers
    assert option2 in all_providers
    assert option3 in all_providers


def test_traverse_typed():
    config = providers.Configuration()
    option = config.option
    typed_option = config.option.as_int()

    all_providers = list(typed_option.traverse())

    assert len(all_providers) == 1
    assert option in all_providers


def test_traverse_overridden():
    options = {"option1": {"option2": "option2"}}
    config = providers.Configuration()
    config.from_dict(options)

    all_providers = list(config.traverse())

    assert len(all_providers) == 1
    overridden, = all_providers
    assert overridden() == options
    assert overridden is config.last_overriding


def test_traverse_overridden_option_1():
    options = {"option2": "option2"}
    config = providers.Configuration()
    config.option1.from_dict(options)

    all_providers = list(config.traverse())

    assert len(all_providers) == 2
    assert config.option1 in all_providers
    assert config.last_overriding in all_providers


def test_traverse_overridden_option_2():
    options = {"option2": "option2"}
    config = providers.Configuration()
    config.option1.from_dict(options)

    all_providers = list(config.option1.traverse())

    assert len(all_providers) == 0



"""Delegate provider traversal tests."""

from dependency_injector import providers


def test_traversal_provider():
    another_provider = providers.Provider()
    provider = providers.Delegate(another_provider)

    all_providers = list(provider.traverse())

    assert len(all_providers) == 1
    assert another_provider in all_providers


def test_traversal_provider_and_overriding():
    provider1 = providers.Provider()
    provider2 = providers.Provider()

    provider3 = providers.Provider()
    provider3.override(provider2)

    provider = providers.Delegate(provider1)

    provider.override(provider3)

    all_providers = list(provider.traverse())

    assert len(all_providers) == 3
    assert provider1 in all_providers
    assert provider2 in all_providers
    assert provider3 in all_providers



"""Dependency provider async mode tests."""

from dependency_injector import providers, errors
from pytest import mark, raises


@mark.asyncio
async def test_provide_error():
    async def get_async():
        raise Exception

    provider = providers.Dependency()
    provider.override(providers.Callable(get_async))

    with raises(Exception):
        await provider()


@mark.asyncio
async def test_isinstance():
    dependency = 1.0

    async def get_async():
        return dependency

    provider = providers.Dependency(instance_of=float)
    provider.override(providers.Callable(get_async))

    assert provider.is_async_mode_undefined() is True

    dependency1 = await provider()

    assert provider.is_async_mode_enabled() is True

    dependency2 = await provider()

    assert dependency1 == dependency
    assert dependency2 == dependency


@mark.asyncio
async def test_isinstance_invalid():
    async def get_async():
        return {}

    provider = providers.Dependency(instance_of=float)
    provider.override(providers.Callable(get_async))

    assert provider.is_async_mode_undefined() is True

    with raises(errors.Error):
        await provider()

    assert provider.is_async_mode_enabled() is True


@mark.asyncio
async def test_async_mode():
    dependency = 123

    async def get_async():
        return dependency

    def get_sync():
        return dependency

    provider = providers.Dependency(instance_of=int)
    provider.override(providers.Factory(get_async))

    assert provider.is_async_mode_undefined() is True

    dependency1 = await provider()

    assert provider.is_async_mode_enabled() is True

    dependency2 = await provider()
    assert dependency1 == dependency
    assert dependency2 == dependency

    provider.override(providers.Factory(get_sync))

    dependency3 = await provider()

    assert provider.is_async_mode_enabled() is True

    dependency4 = await provider()
    assert dependency3 == dependency
    assert dependency4 == dependency



"""Container traversing tests."""

from dependency_injector import containers, providers


class Container(containers.DeclarativeContainer):
    obj_factory = providers.DelegatedFactory(
        dict,
        foo=providers.Resource(
            dict,
            foo="bar"
        ),
        bar=providers.Resource(
            dict,
            foo="bar"
        )
    )


def test_nested_providers():
    container = Container()
    all_providers = list(container.traverse())

    assert container.obj_factory in all_providers
    assert container.obj_factory.kwargs["foo"] in all_providers
    assert container.obj_factory.kwargs["bar"] in all_providers
    assert len(all_providers) == 3


def test_nested_providers_with_filtering():
    container = Container()
    all_providers = list(container.traverse(types=[providers.Resource]))

    assert container.obj_factory.kwargs["foo"] in all_providers
    assert container.obj_factory.kwargs["bar"] in all_providers
    assert len(all_providers) == 2


def test_container_cls_nested_providers():
    all_providers = list(Container.traverse())

    assert Container.obj_factory in all_providers
    assert Container.obj_factory.kwargs["foo"] in all_providers
    assert Container.obj_factory.kwargs["bar"] in all_providers
    assert len(all_providers) == 3


def test_container_cls_nested_providers_with_filtering():
    all_providers = list(Container.traverse(types=[providers.Resource]))

    assert Container.obj_factory.kwargs["foo"] in all_providers
    assert Container.obj_factory.kwargs["bar"] in all_providers
    assert len(all_providers) == 2



from typing import Dict

from dependency_injector import containers, providers


# Test 1: to check setattr
container1 = containers.DynamicContainer()
container1.abc = providers.Provider()

# Test 2: to check override()
container2 = containers.DynamicContainer()
container2.override(containers.DynamicContainer())

# Test 3: to check override_providers()
container3 = containers.DynamicContainer()
container3.override_providers(a=providers.Provider())
with container3.override_providers(a=providers.Provider()):
    ...

# Test 4: to check set_providers()
container4 = containers.DynamicContainer()
container4.set_providers(a=providers.Provider())

# Test 5: to check .dependencies attribute
container5 = containers.DynamicContainer()
dependencies: Dict[str, providers.Provider] = container5.dependencies

# Test 6: to check base class
container6: containers.Container = containers.DynamicContainer()



"""DependencyContainer provider tests."""

from dependency_injector import containers, providers, errors
from pytest import fixture, raises


class Container(containers.DeclarativeContainer):

    dependency = providers.Provider()


@fixture
def provider():
    return providers.DependenciesContainer()


@fixture
def container():
    return Container()


def test_getattr(provider):
    has_dependency = hasattr(provider, "dependency")
    dependency = provider.dependency

    assert isinstance(dependency, providers.Dependency)
    assert dependency is provider.dependency
    assert has_dependency is True
    assert dependency.last_overriding is None


def test_getattr_with_container(provider, container):
    provider.override(container)

    dependency = provider.dependency

    assert dependency.overridden == (container.dependency,)
    assert dependency.last_overriding is container.dependency


def test_providers(provider):
    dependency1 = provider.dependency1
    dependency2 = provider.dependency2
    assert provider.providers == {"dependency1": dependency1, "dependency2": dependency2}


def test_override(provider, container):
    dependency = provider.dependency
    provider.override(container)

    assert dependency.overridden == (container.dependency,)
    assert dependency.last_overriding is container.dependency


def test_reset_last_overriding(provider, container):
    dependency = provider.dependency
    provider.override(container)
    provider.reset_last_overriding()

    assert dependency.last_overriding is None
    assert dependency.last_overriding is None


def test_reset_override(provider, container):
    dependency = provider.dependency
    provider.override(container)
    provider.reset_override()

    assert dependency.overridden == tuple()
    assert not dependency.overridden


def test_assign_parent(provider):
    parent = providers.DependenciesContainer()
    provider.assign_parent(parent)
    assert provider.parent is parent


def test_parent_name(provider):
    container = containers.DynamicContainer()
    container.name = provider
    assert provider.parent_name == "name"


def test_parent_name_with_deep_parenting(provider):
    container = providers.DependenciesContainer(name=provider)
    _ = providers.DependenciesContainer(container=container)
    assert provider.parent_name == "container.name"


def test_parent_name_is_none(provider):
    assert provider.parent_name is None


def test_parent_deepcopy(provider, container):
    container.name = provider
    copied = providers.deepcopy(container)

    assert container.name.parent is container
    assert copied.name.parent is copied

    assert container is not copied
    assert container.name is not copied.name
    assert container.name.parent is not copied.name.parent


def test_parent_set_on__getattr__(provider):
    assert isinstance(provider.name, providers.Dependency)
    assert provider.name.parent is provider


def test_parent_set_on__init__():
    provider = providers.Dependency()
    container = providers.DependenciesContainer(name=provider)
    assert container.name is provider
    assert container.name.parent is container


def test_resolve_provider_name(provider):
    assert provider.resolve_provider_name(provider.name) == "name"


def test_resolve_provider_name_no_provider(provider):
    with raises(errors.Error):
        provider.resolve_provider_name(providers.Provider())



"""AbstractFactory provider tests."""

from dependency_injector import providers, errors
from pytest import raises

from .common import Example


def test_inheritance():
    assert isinstance(providers.AbstractFactory(Example), providers.Factory)


def test_call_overridden_by_factory():
    provider = providers.AbstractFactory(object)
    provider.override(providers.Factory(Example))
    assert isinstance(provider(), Example)


def test_call_overridden_by_delegated_factory():
    provider = providers.AbstractFactory(object)
    provider.override(providers.DelegatedFactory(Example))
    assert isinstance(provider(), Example)


def test_call_not_overridden():
    provider = providers.AbstractFactory(object)
    with raises(errors.Error):
        provider()


def test_override_by_not_factory():
    provider = providers.AbstractFactory(object)
    with raises(errors.Error):
        provider.override(providers.Callable(object))


def test_provide_not_implemented():
    provider = providers.AbstractFactory(Example)
    with raises(NotImplementedError):
        provider._provide(tuple(), dict())


def test_repr():
    provider = providers.AbstractFactory(Example)
    assert repr(provider) == (
        "<dependency_injector.providers."
        "AbstractFactory({0}) at {1}>".format(repr(Example), hex(id(provider)))
    )



"""Configuration provider tests."""

import decimal

from dependency_injector import containers, providers, errors
from pytest import mark, raises


def test_init_optional(config):
    config.set_name("myconfig")
    config.set_default({"foo": "bar"})
    config.set_strict(True)

    assert config.get_name() == "myconfig"
    assert config.get_default() == {"foo": "bar"}
    assert config.get_strict() is True


def test_set_name_returns_self(config):
    assert config.set_name("myconfig") is config


def test_set_default_returns_self(config):
    assert config.set_default({}) is config


def test_set_strict_returns_self(config):
    assert config.set_strict(True) is config


def test_default_name(config):
    assert config.get_name() == "config"


def test_providers_are_providers(config):
    assert providers.is_provider(config.a) is True
    assert providers.is_provider(config.a.b) is True
    assert providers.is_provider(config.a.b.c) is True
    assert providers.is_provider(config.a.b.d) is True


def test_providers_are_not_delegates(config):
    assert providers.is_delegated(config.a) is False
    assert providers.is_delegated(config.a.b) is False
    assert providers.is_delegated(config.a.b.c) is False
    assert providers.is_delegated(config.a.b.d) is False


def test_providers_identity(config):
    assert config.a is config.a
    assert config.a.b is config.a.b
    assert config.a.b.c is config.a.b.c
    assert config.a.b.d is config.a.b.d


def test_get_name(config):
    assert config.a.b.c.get_name() == "config.a.b.c"


def test_providers_value_setting(config):
    a = config.a
    ab = config.a.b
    abc = config.a.b.c
    abd = config.a.b.d

    config.update({"a": {"b": {"c": 1, "d": 2}}})

    assert a() == {"b": {"c": 1, "d": 2}}
    assert ab() == {"c": 1, "d": 2}
    assert abc() == 1
    assert abd() == 2


def test_providers_with_already_set_value(config):
    config.update({"a": {"b": {"c": 1, "d": 2}}})

    a = config.a
    ab = config.a.b
    abc = config.a.b.c
    abd = config.a.b.d

    assert a() == {"b": {"c": 1, "d": 2}}
    assert ab() == {"c": 1, "d": 2}
    assert abc() == 1
    assert abd() == 2


def test_as_int(config):
    value_provider = providers.Callable(lambda value: value, config.test.as_int())
    config.from_dict({"test": "123"})

    value = value_provider()
    assert value == 123


def test_as_float(config):
    value_provider = providers.Callable(lambda value: value, config.test.as_float())
    config.from_dict({"test": "123.123"})

    value = value_provider()
    assert value == 123.123


def test_as_(config):
    value_provider = providers.Callable(
        lambda value: value,
        config.test.as_(decimal.Decimal),
    )
    config.from_dict({"test": "123.123"})

    value = value_provider()
    assert value == decimal.Decimal("123.123")


def test_required(config):
    provider = providers.Callable(
        lambda value: value,
        config.a.required(),
    )
    with raises(errors.Error, match="Undefined configuration option \"config.a\""):
        provider()


def test_required_defined_none(config):
    provider = providers.Callable(
        lambda value: value,
        config.a.required(),
    )
    config.from_dict({"a": None})
    assert provider() is None


def test_required_no_side_effect(config):
    _ = providers.Callable(
        lambda value: value,
        config.a.required(),
    )
    assert config.a() is None


def test_required_as_(config):
    provider = providers.List(
        config.int_test.required().as_int(),
        config.float_test.required().as_float(),
        config._as_test.required().as_(decimal.Decimal),
    )
    config.from_dict({"int_test": "1", "float_test": "2.0", "_as_test": "3.0"})
    assert provider() == [1, 2.0, decimal.Decimal("3.0")]


def test_providers_value_override(config):
    a = config.a
    ab = config.a.b
    abc = config.a.b.c
    abd = config.a.b.d

    config.override({"a": {"b": {"c": 1, "d": 2}}})

    assert a() == {"b": {"c": 1, "d": 2}}
    assert ab() == {"c": 1, "d": 2}
    assert abc() == 1
    assert abd() == 2


def test_configuration_option_override_and_reset_override(config):
    # Bug: https://github.com/ets-labs/python-dependency-injector/issues/319
    config.from_dict({"a": {"b": {"c": 1}}})

    assert config.a.b.c() == 1

    with config.set("a.b.c", "xxx"):
        assert config.a.b.c() == "xxx"
    assert config.a.b.c() == 1

    with config.a.b.c.override("yyy"):
        assert config.a.b.c() == "yyy"

    assert config.a.b.c() == 1


def test_providers_with_already_overridden_value(config):
    config.override({"a": {"b": {"c": 1, "d": 2}}})

    a = config.a
    ab = config.a.b
    abc = config.a.b.c
    abd = config.a.b.d

    assert a() == {"b": {"c": 1, "d": 2}}
    assert ab() == {"c": 1, "d": 2}
    assert abc() == 1
    assert abd() == 2


def test_providers_with_default_value(config):
    config.set_default({"a": {"b": {"c": 1, "d": 2}}})

    a = config.a
    ab = config.a.b
    abc = config.a.b.c
    abd = config.a.b.d

    assert a() == {"b": {"c": 1, "d": 2}}
    assert ab() == {"c": 1, "d": 2}
    assert abc() == 1
    assert abd() == 2


def test_providers_with_default_value_overriding(config):
    config.set_default({"a": {"b": {"c": 1, "d": 2}}})

    assert config.a() == {"b": {"c": 1, "d": 2}}
    assert config.a.b() == {"c": 1, "d": 2}
    assert config.a.b.c() == 1
    assert config.a.b.d() == 2

    config.override({"a": {"b": {"c": 3, "d": 4}}})
    assert config.a() == {"b": {"c": 3, "d": 4}}
    assert config.a.b() == {"c": 3, "d": 4}
    assert config.a.b.c() == 3
    assert config.a.b.d() == 4

    config.reset_override()
    assert config.a() == {"b": {"c": 1, "d": 2}}
    assert config.a.b() == {"c": 1, "d": 2}
    assert config.a.b.c() == 1
    assert config.a.b.d() == 2


def test_value_of_undefined_option(config):
    assert config.option() is None


@mark.parametrize("config_type", ["strict"])
def test_value_of_undefined_option_in_strict_mode(config):
    with raises(errors.Error, match="Undefined configuration option \"config.option\""):
        config.option()


@mark.parametrize("config_type", ["strict"])
def test_value_of_undefined_option_with_root_none_in_strict_mode(config):
    config.override(None)
    with raises(errors.Error, match="Undefined configuration option \"config.option\""):
        config.option()


@mark.parametrize("config_type", ["strict"])
def test_value_of_defined_none_option_in_strict_mode(config):
    config.from_dict({"a": None})
    assert config.a() is None


def test_getting_of_special_attributes(config):
    with raises(AttributeError):
        config.__name__


def test_getting_of_special_attributes_from_child(config):
    with raises(AttributeError):
        config.child.__name__


def test_context_manager_alias():
    class Container(containers.DeclarativeContainer):
        config = providers.Configuration()

    container = Container()

    with container.config as config:
        config.override({"foo": "foo", "bar": "bar"})

    assert container.config() == {"foo": "foo", "bar": "bar"}
    assert config() == {"foo": "foo", "bar": "bar"}
    assert container.config is config


def test_option_context_manager_alias():
    class Container(containers.DeclarativeContainer):
        config = providers.Configuration()

    container = Container()

    with container.config.option as option:
        option.override({"foo": "foo", "bar": "bar"})

    assert container.config() == {"option": {"foo": "foo", "bar": "bar"}}
    assert container.config.option() == {"foo": "foo", "bar": "bar"}
    assert option() == {"foo": "foo", "bar": "bar"}
    assert container.config.option is option


def test_missing_key(config):
    # See: https://github.com/ets-labs/python-dependency-injector/issues/358
    config.override(None)
    value = config.key()
    assert value is None


def test_deepcopy(config):
    config_copy = providers.deepcopy(config)
    assert isinstance(config_copy, providers.Configuration)
    assert config is not config_copy


def test_deepcopy_from_memo(config):
    config_copy_memo = providers.Configuration()

    provider_copy = providers.deepcopy(config, memo={id(config): config_copy_memo})
    assert provider_copy is config_copy_memo


def test_deepcopy_overridden(config):
    object_provider = providers.Object(object())

    config.override(object_provider)

    provider_copy = providers.deepcopy(config)
    object_provider_copy = provider_copy.overridden[0]

    assert config is not provider_copy
    assert isinstance(config, providers.Configuration)

    assert object_provider is not object_provider_copy
    assert isinstance(object_provider_copy, providers.Object)


def test_repr(config):
    assert repr(config) == (
        "<dependency_injector.providers."
        "Configuration({0}) at {1}>".format(repr("config"), hex(id(config)))
    )


def test_repr_child(config):
    assert repr(config.a.b.c) == (
        "<dependency_injector.providers."
        "ConfigurationOption({0}) at {1}>".format(repr("config.a.b.c"), hex(id(config.a.b.c)))
    )



"""Singleton provider async mode tests."""

import asyncio
import random

from dependency_injector import providers
from pytest import mark, raises

from .common import RESOURCE1, RESOURCE2, BaseContainer, Client, Service


@mark.asyncio
async def test_injections():
    class ContainerWithSingletons(BaseContainer):
        client = providers.Singleton(
            Client,
            resource1=BaseContainer.resource1,
            resource2=BaseContainer.resource2,
        )

        service = providers.Singleton(
            Service,
            client=client,
        )

    container = ContainerWithSingletons()

    client1 = await container.client()
    client2 = await container.client()

    assert isinstance(client1, Client)
    assert client1.resource1 is RESOURCE1
    assert client1.resource2 is RESOURCE2

    assert isinstance(client2, Client)
    assert client2.resource1 is RESOURCE1
    assert client2.resource2 is RESOURCE2

    service1 = await container.service()
    service2 = await container.service()

    assert isinstance(service1, Service)
    assert isinstance(service1.client, Client)
    assert service1.client.resource1 is RESOURCE1
    assert service1.client.resource2 is RESOURCE2

    assert isinstance(service2, Service)
    assert isinstance(service2.client, Client)
    assert service2.client.resource1 is RESOURCE1
    assert service2.client.resource2 is RESOURCE2

    assert service1 is service2
    assert service1.client is service2.client
    assert service1.client is client1

    assert service2.client is client2
    assert client1 is client2


@mark.asyncio
async def test_async_mode():
    instance = object()

    async def create_instance():
        return instance

    provider = providers.Singleton(create_instance)

    instance1 = await provider()
    instance2 = await provider()

    assert instance1 is instance2
    assert instance1 is instance
    assert instance2 is instance


@mark.asyncio
async def test_concurrent_init():
    async def create_instance():
        await asyncio.sleep(random.randint(1, 10) / 1000)
        return object()

    provider = providers.Singleton(create_instance)

    future_instance1 = provider()
    future_instance2 = provider()

    instance1, instance2 = await asyncio.gather(future_instance1, future_instance2)
    instance3 = await provider()

    assert instance1 is instance2 is instance3


@mark.asyncio
async def test_async_init_with_error():
    async def create_instance():
        create_instance.counter += 1
        raise RuntimeError()

    create_instance.counter = 0

    provider = providers.Singleton(create_instance)

    future = provider()
    assert provider.is_async_mode_enabled() is True

    with raises(RuntimeError):
        await future

    assert create_instance.counter == 1
    assert provider.is_async_mode_enabled() is True

    with raises(RuntimeError):
        await provider()

    assert create_instance.counter == 2
    assert provider.is_async_mode_enabled() is True



"""DelegatedFactory provider tests."""

from dependency_injector import providers

from .common import Example


def test_inheritance():
    assert isinstance(providers.DelegatedFactory(object), providers.Factory)


def test_is_provider():
    assert providers.is_provider(providers.DelegatedFactory(object)) is True


def test_is_delegated_provider():
    assert providers.is_delegated(providers.DelegatedFactory(object)) is True


def test_repr():
    provider = providers.DelegatedFactory(Example)
    assert repr(provider) == (
        "<dependency_injector.providers."
        "DelegatedFactory({0}) at {1}>".format(repr(Example), hex(id(provider)))
    )



"""Configuration.from_yaml() tests."""

from dependency_injector import providers, errors
from pytest import fixture, mark, raises


@fixture
def no_yaml_module_installed():
    yaml = providers.yaml
    providers.yaml = None
    yield
    providers.yaml = yaml


def test(config, yaml_config_file_1):
    config.from_yaml(yaml_config_file_1)

    assert config() == {"section1": {"value1": 1}, "section2": {"value2": 2}}
    assert config.section1() == {"value1": 1}
    assert config.section1.value1() == 1
    assert config.section2() == {"value2": 2}
    assert config.section2.value2() == 2


def test_merge(config, yaml_config_file_1, yaml_config_file_2):
    config.from_yaml(yaml_config_file_1)
    config.from_yaml(yaml_config_file_2)

    assert config() == {
        "section1": {
            "value1": 11,
            "value11": 11,
        },
        "section2": {
            "value2": 2,
        },
        "section3": {
            "value3": 3,
        },
    }
    assert config.section1() == {"value1": 11, "value11": 11}
    assert config.section1.value1() == 11
    assert config.section1.value11() == 11
    assert config.section2() == {"value2": 2}
    assert config.section2.value2() == 2
    assert config.section3() == {"value3": 3}
    assert config.section3.value3() == 3


def test_file_does_not_exist(config):
    config.from_yaml("./does_not_exist.yml")
    assert config() == {}


@mark.parametrize("config_type", ["strict"])
def test_file_does_not_exist_strict_mode(config):
    with raises(IOError):
        config.from_yaml("./does_not_exist.yml")


def test_option_file_does_not_exist(config):
    config.option.from_yaml("./does_not_exist.yml")
    assert config.option() is None


@mark.parametrize("config_type", ["strict"])
def test_option_file_does_not_exist_strict_mode(config):
    with raises(IOError):
        config.option.from_yaml("./does_not_exist.yml")


def test_required_file_does_not_exist(config):
    with raises(IOError):
        config.from_yaml("./does_not_exist.yml", required=True)


def test_required_option_file_does_not_exist(config):
    with raises(IOError):
        config.option.from_yaml("./does_not_exist.yml", required=True)


@mark.parametrize("config_type", ["strict"])
def test_not_required_file_does_not_exist_strict_mode(config):
    config.from_yaml("./does_not_exist.yml", required=False)
    assert config() == {}


@mark.parametrize("config_type", ["strict"])
def test_not_required_option_file_does_not_exist_strict_mode(config):
    config.option.from_yaml("./does_not_exist.yml", required=False)
    with raises(errors.Error):
        config.option()


@mark.usefixtures("no_yaml_module_installed")
def test_no_yaml_installed(config, yaml_config_file_1):
    with raises(errors.Error) as error:
        config.from_yaml(yaml_config_file_1)
    assert error.value.args[0] == (
        "Unable to load yaml configuration - PyYAML is not installed. "
        "Install PyYAML or install Dependency Injector with yaml extras: "
        "\"pip install dependency-injector[yaml]\""
    )


@mark.usefixtures("no_yaml_module_installed")
def test_option_no_yaml_installed(config, yaml_config_file_1):
    with raises(errors.Error) as error:
        config.option.from_yaml(yaml_config_file_1)
    assert error.value.args[0] == (
        "Unable to load yaml configuration - PyYAML is not installed. "
        "Install PyYAML or install Dependency Injector with yaml extras: "
        "\"pip install dependency-injector[yaml]\""
    )



"""Dependency injector errors."""


class Error(Exception):
    """Base error.

    All dependency injector errors extend this error class.
    """


class NoSuchProviderError(Error, AttributeError):
    """Error that is raised when provider lookup is failed."""



from typing import Coroutine

from dependency_injector import providers


async def _coro() -> None:
    ...

# Test 1: to check the return type
provider1 = providers.Coroutine(_coro)
var1: Coroutine = provider1()

# Test 2: to check string imports
provider2: providers.Coroutine[None] = providers.Coroutine("_coro")
provider2.set_provides("_coro")
